#!/usr/bin/python

import os, sys, poplib, imaplib, subprocess, multiprocessing, logging, socket, random, time
from pprint import pprint
from optparse import OptionParser

# Deal with commandline options
parser = OptionParser()
parser.add_option( '-l', '--log', dest='logfile', help='What file to log to', metavar='FILE', default='/var/log/dovecot/poolmon/poolmon.log' )
parser.add_option( '-c', '--credfile', dest='credfile', help='File that contains login info', metavar='FILE', default='/etc/dovecot/poolmon.creds' )
parser.add_option( '-t', '--timeout', dest='timeout', help='Timeout for pop/imap connections', default=10 )
parser.add_option( '-d', '--debug', dest='debug', action='store_true', help='Turn on debugging', )
parser.add_option( '-s', '--nosleep', dest='nosleep', action='store_true', help='Do not sleep after start', )
parser.add_option( '-r', '--retry', dest='retry', default=1, help='Number of retries', )
parser.add_option( '-f', '--force', dest='force', action='store_true', help='Run no matter what', )
( options, args ) = parser.parse_args()

# Turn on debugging if asked
if options.debug:
	logging.basicConfig( format='%(asctime)s %(message)s', filename=options.logfile, level=logging.DEBUG )
else:
	logging.basicConfig( format='%(asctime)s %(message)s', filename=options.logfile, level=logging.INFO )


# Set global timeout (imaplib doesn't let you set one)
socket.setdefaulttimeout( int( options.timeout ) )


# Connect to POP and log in
def check_pop( hostname, username, password, timeout=10 ):
	# Debug
	log( "Connecting to host %s via POP3 with timeout %d" % ( hostname, timeout ), logging.DEBUG )

	# Try to connect to POP
	try:
		conn = poplib.POP3( hostname, timeout=timeout )
	except Exception, e:
		log( "Can't connect to host %s via POP3: %s" % ( hostname, e ), logging.ERROR )
		return

	# Try to auth to POP
	try:
		conn.user( username )
		conn.pass_( password )
	except Exception, e:
		log( "Can't log into to host %s via POP3: %s" % ( hostname, e ), logging.ERROR )
		return

	# Disconnect
	conn.quit()

	return True


# Connect to IMAP and log in
def check_imap( hostname, username, password, ):
	# Debug
	log( "Connecting to host %s via IMAP" % ( hostname, ), logging.DEBUG )

	# Try to connect to IMAP
	try:
		conn = imaplib.IMAP4( hostname, )
	except Exception, e:
		# Debug
		log( "Can't connect to host %s via IMAP: %s" % ( hostname, e ), logging.ERROR )
		return

	# Try to auth to IMAP
	try:
		conn.login( username, password )
	except Exception, e:
		# Debug
		log( "Can't log into to host %s via IMAP: %s" % ( hostname, e ), logging.ERROR )
		return

	# Disconnect
	conn.logout()

	return True


# Read credentials file
def read_creds( filename ):
	with open( filename, 'r' ) as fh:
		lines = fh.readlines()
		username = lines[0].strip()
		password = lines[1].strip()
		return username, password


# Get servers and parse status
def get_directors():
	servers = {}
	ip_me, ip_left = None, None

	# Run status command
	subp = subprocess.Popen( [ 'doveadm', '-f', 'flow', 'director', 'ring', 'status', ], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE )

	# Dump stderr
	_ = subp.stderr.read()

	for line in subp.stdout.readlines():
		comps = line.split()

		if not comps[-1] == 'status=synced': continue

		if not comps[3] in ( 'type=self', 'type=left', ): continue

		_, ip = comps[1].split( "=" )

		if comps[3] == 'type=self':
			ip_me = ip_to_int( ip )
		elif comps[3] == 'type=left':
			ip_left = ip_to_int( ip )

	return ip_me, ip_left


def lowest_director( left, me ):
	if ip_me < ip_left:
		return True
	else:
		return False


def ip_to_int( ip ):
	comps = ip.split( "." )
	return ( int( comps[0] ) << 24 ) + ( int( comps[1] ) << 16 ) + ( int( comps[2] ) << 8 ) + int( comps[3] )


# Get servers and parse status
def get_servers():
	servers = {}

	# Run status command
	subp = subprocess.Popen( [ 'doveadm', '-f', 'flow', 'director', 'status', ], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE )

	# Dump stderr
	_ = subp.stderr.read()

	for line in subp.stdout.readlines():
		server, weight, state = None, None, None

		comps = line.strip().split()

		for comp in comps:
			if comp.startswith( "ip=" ):
				( _, server ) = comp.split( "=" )
			elif comp.startswith( "vhosts=" ):
				( _, weight ) = comp.split( "=" )
			elif comp.startswith( "state=" ):
				( _, state ) = comp.split( "=" )

		if state == 'down' or int( weight ) < 1:
			servers[ server ] = 'down'

			# Corner case, so toggle it too
			if ( state == 'up' and int( weight ) < 1 ) or ( state == 'down' and int( weight ) > 0 ):
				log( "Host %s in weird state so setting 'down'" % ( server ), logging.DEBUG )
				toggle_host( server, 'down' )
				time.sleep(1)
				set_host_weight( server, '0' )
				time.sleep(1)
				flush_host( server )
		else:
			servers[ server ] = 'up'

	return servers


# Logging function
def log( message, loglevel ):
	logfunc = logging.info

	if loglevel == logging.DEBUG:
		logfunc = logging.debug

	logfunc( " [%d] %s" % ( os.getpid(), message ) )


# Main process for workers
def worker( hostname, username, password, state, timeout=10, ):

	host_ok = True

	# Debug
	log( "Hostname: %s" % ( hostname ), logging.DEBUG )
	log( "Username: %s" % ( username ), logging.DEBUG )
	log( "Password: %s" % ( password ), logging.DEBUG )
	log( "Timeout:  %d" % ( timeout ), logging.DEBUG )

	# Try up to '--retry' times to check server. If we get a good response, stop loop
	for i in range( 0, int( options.retry ) ):
		host_ok = try_server( hostname, username, password, state, timeout )
		log( "Try %d for host %s: %s" % ( i+1, hostname, host_ok ), logging.DEBUG )

		if host_ok:
			break
		else:
			time.sleep(1)


	# If host failed, set 'down' if it wasn't already
	if not host_ok:
		if state != 'down':
			# Debug
			log( "Setting host 'down': %s" % ( hostname ), logging.WARN )
			toggle_host( hostname, 'down' )
			time.sleep(1)
			set_host_weight( hostname, '0' )
			time.sleep(1)
			flush_host( hostname )
		else:
			log( "Host already 'down': %s (%s)" % ( hostname, state ), logging.WARN )

	# If host is ok, set 'up' (and set weight) if it wasn't already
	else:
		if state != 'up':
			# Debug
			log( "Setting host 'up': %s" % ( hostname ), logging.WARN )
			toggle_host( hostname, 'up' )
			time.sleep(1)
			set_host_weight( hostname, '100' )
		else:
			# Debug
			log( "Host already 'up': %s" % ( hostname ), logging.DEBUG )


def try_server( hostname, username, password, state, timeout=10, ):
	host_ok = True

	# Run POP check
	pop_rv = check_pop( hostname, username, password, timeout )

	# If we failed POP check, don't bother with IMAP
	if not pop_rv:
		# Debug
		log( "Host not ok via POP3: %s" % ( hostname ), logging.INFO )
		host_ok = False

	# Run IMAP check
	if pop_rv:
		imap_rv = check_imap( hostname, username, password, )

		# If IMAP check failed
		if not imap_rv:
			# Debug
			log( "Host not ok via IMAP: %s" % ( hostname ), logging.INFO )
			host_ok = False

	return host_ok


# Set weight for host: If host is 'up' but weight is 0, it won't be used
def set_host_weight( hostname, weight ):
	subp = subprocess.Popen( [ 'doveadm', 'director', 'update', hostname, weight ], stdout=subprocess.PIPE, stderr=subprocess.PIPE )

	_ = subp.stderr.read()
	_ = subp.stdout.read()


# Set host to 'down' or 'up'
def toggle_host( hostname, action ):
	subp = subprocess.Popen( [ 'doveadm', 'director', action, hostname ], stdout=subprocess.PIPE, stderr=subprocess.PIPE )

	_ = subp.stderr.read()
	_ = subp.stdout.read()


# Flush hosts from down server
def flush_host( hostname ):
	subp = subprocess.Popen( [ 'doveadm', 'director', 'flush', hostname ], stdout=subprocess.PIPE, stderr=subprocess.PIPE )

	_ = subp.stderr.read()
	_ = subp.stdout.read()


# My IP and the 'left' IP
ip_me, ip_left = get_directors()

# Unless forcing, quit if IP isn't lowest director (we don't want everyone doing checks)
if not options.force:
	if not lowest_director( ip_left, ip_me ):
		log( "Not the lowest-numbered director, so quitting", logging.INFO )
		sys.exit()


# Sleep if not told otherwise
if not options.nosleep:
	time.sleep( random.randint( 0, 30 ) )


# Get dict of servers with value of 'up' or 'down'
servers = get_servers()


# Read credential file
username, password = read_creds( options.credfile )


# Run
if __name__ == '__main__':
	jobs = []

	# Walk servers dict and spin off a proc for each
	for hostname in servers.keys():
		p = multiprocessing.Process(target=worker, args=( hostname, username, password, servers[hostname], int( options.timeout ), ) )
		jobs.append(p)
		p.start()

